/* ui.c:

   Contains wmcd's user interface

   */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

#include "xpm/wmcd-shaped.xpm"
#include "xpm/wmcd.xpm"
#include "xpm/led.xpm"
#include "xpm/opentray.xpm"
#include "xpm/opentray-p.xpm"
#include "xpm/closetray.xpm"
#include "xpm/closetray-p.xpm"
#include "xpm/music.xpm"
#include "xpm/data.xpm"
#include "xpm/ctls/play.xpm"
#include "xpm/ctls/play-p.xpm"
#include "xpm/ctls/play-i.xpm"
#include "xpm/ctls/stop.xpm"
#include "xpm/ctls/stop-p.xpm"
#include "xpm/ctls/stop-i.xpm"
#include "xpm/ctls/pause.xpm"
#include "xpm/ctls/pause-p.xpm"
#include "xpm/ctls/pause-i.xpm"
#include "xpm/ctls/skipf.xpm"
#include "xpm/ctls/skipf-p.xpm"
#include "xpm/ctls/skipf-i.xpm"
#include "xpm/ctls/skipb.xpm"
#include "xpm/ctls/skipb-p.xpm"
#include "xpm/ctls/skipb-i.xpm"

#include "xpm/msgs/startup.xpm"
#include "xpm/msgs/mounted.xpm"

#include "cdrom.h"

#include "typedefs.h"
#include "actions.h"
#include "window.h"
#include "config.h"

Display *dpy;
Window root;
int screen;
int d_depth;

Pixel back_pix, fore_pix;
GC NormalGC;
Window iconwin, win, theWin;       /* My home is my window */
Cursor stopwatch;

#define MW_EVENTS   (ExposureMask | ButtonPressMask |\
		     ButtonReleaseMask | StructureNotifyMask)

/* XPM-Stuff */

XpmIcon wmcd, led;
XpmIcon message;
XpmIcon trayctl, trayctl_p, opentray, opentray_p, closetray, closetray_p;
XpmIcon music, data;

XpmIcon play, play_a, play_i, play_p;
XpmIcon stop, stop_a, stop_i, stop_p;
XpmIcon pause_img, pause_a, pause_i, pause_p;
XpmIcon skipf, skipf_a, skipf_i, skipf_p;
XpmIcon skipb, skipb_a, skipb_i, skipb_p;

/*
 * state variables
 */

extern int play_active;
extern int stop_active;
extern int pause_active;
extern int skipf_active;
extern int skipb_active;


/* Die Positionen: */
#define NOSHAPE 0
#define SHAPE   1

int shape = NOSHAPE;

#define TRAYCTL  0
#define MESSAGE  1
#define TRACK    2
#define POS      3
#define PAUSE    4
#define TRACKLED 5
#define POSLED_M 6
#define POSLED_S 7
#define CONTROLS 8
#define M_D_ICON 9

int xpos[10] = {18,  4,  5, 25, 56, 14, 34, 46,  2,  5};
int ypos[10] = { 3, 36, 36, 36, 36, 36, 36, 36, 50, 20};



/* button regions */

#define TRAYCTL_R 0
#define PLAY_R    1
#define STOP_R    2
#define PAUSE_R   3
#define SKIPF_R   4
#define SKIPB_R   5

Region regions[6];



#define LEDWIDTH 5
#define LEDOFFSET 22
#define LEDHEIGHT 10

#define CUSTOMWIDTH 8 /* "Custom Display" symbols */
#define TKOFFSET 0   
#define PSOFFSET 8

#define PAUSEOFFSET 16 /* the small "pause" icon */
#define PAUSEWIDTH 4 

#define COLONOFFSET 20
#define COLONWIDTH 2

#define CONTROLWIDTH  12 /* The controls */
#define CONTROLHEIGHT 12 

#define M_D_HEIGHT 10 /* The music/data icon */
#define M_D_WIDTH   7
#define M_D_SKIP    2

extern char *ERR_colorcells;

extern int show_led; /* This tells us whether to show a position LED
			or a predefined message (MOUNTED etc.) */

extern int track_playing;
extern cdrom_position_t position;

int shaped = True;
int iconic = True;

static int l_play = False, l_stop = False, l_pause = False,
  l_skipf = False, l_skipb = False;
static int l_tray = False;
static int l_music = False;
static int l_data = False;
static int l_mounted = False;

/* function prototypes */

void EvalState(void);
int ReadPosition(void); /* returns 1 if the display ought to be updated */

/**********************************************************************
 *
 *          Misc
 *
 **********************************************************************/


/*
 * Try to allocate the named color
 */

Pixel GetColor(char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes(dpy, root, &attributes);
  color.pixel = 0;
   if (!XParseColor (dpy, attributes.colormap, name, &color)) 
     {
       fprintf(stderr, "asclock: Can't parse color %s\n", name);
     }
   else if(!XAllocColor (dpy, attributes.colormap, &color)) 
     {
       fprintf(stderr, "asclock: Can't alloc color %s\n",name);
     }
  return color.pixel;
}


/*
 * Flush all expose events out of the event queue
 */

int FlushExpose (Window w)
{
  XEvent dummy;
  int i=0;
  
  while (XCheckTypedWindowEvent (dpy, w, Expose, &dummy))i++;
  return i;
}



/**********************************************************************
 *
 *          The LED display
 *
 **********************************************************************/


/*
 * draw the position LED
 */

void DrawLED(void)
{

  int n;

  if(!show_led)
    return;
  
  XSetClipMask(dpy, NormalGC, message.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[MESSAGE], ypos[MESSAGE]);
  XFillRectangle(dpy, theWin, NormalGC, xpos[MESSAGE],
		 ypos[MESSAGE], message.attributes.width,
		 message.attributes.height);
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, TKOFFSET, 0, CUSTOMWIDTH,
	    LEDHEIGHT, xpos[TRACK], ypos[TRACK]);
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, PSOFFSET, 0, CUSTOMWIDTH,
	    LEDHEIGHT, xpos[POS], ypos[POS]);
  if(cdrom_paused())
    XCopyArea(dpy, led.pixmap, theWin, NormalGC, PAUSEOFFSET, 0, PAUSEWIDTH,
	      LEDHEIGHT, xpos[PAUSE], ypos[PAUSE]);
  
  /* Track display */

  if(track_playing == -1)
    {
      track_playing = cdrom_current_track();
      /*      while(hw->tocentries[track_playing-1].cdte_ctrl ==
	    CDROM_DATA_TRACK)
	{
	  printf("Track %d is DATA\n", track_playing);
	  track_playing++;
	  if(track_playing > hw->tochdr.cdth_trk1)
	    {
	      fprintf(stderr, "Inconsistency in data/audio track database\n");
	      exit(1);
	    }
	}
	*/
    }
  n = track_playing / 10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[TRACKLED], ypos[TRACKLED]);
  n = track_playing % 10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[TRACKLED]+LEDWIDTH, ypos[TRACKLED]);
  /* MM:SS display */
  n = position.minute/10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[POSLED_M], ypos[POSLED_M]);
  n = position.minute%10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[POSLED_M]+LEDWIDTH,
	    ypos[POSLED_M]);


  XCopyArea(dpy, led.pixmap, theWin, NormalGC, COLONOFFSET, 0,
	    COLONWIDTH, LEDHEIGHT, xpos[POSLED_M]+2*LEDWIDTH,
	    ypos[POSLED_M]);

  n = position.second/10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[POSLED_S], ypos[POSLED_S]);
  n = position.second%10;
  XCopyArea(dpy, led.pixmap, theWin, NormalGC, LEDOFFSET+n*LEDWIDTH, 0, LEDWIDTH,
	    LEDHEIGHT, xpos[POSLED_S]+LEDWIDTH, ypos[POSLED_S]);
  
}


/*
 * Create a message pixmap from bytes
 */

void CreateMessage(char **bytes)
{
  message.attributes.valuemask |= (XpmReturnPixels |
				   XpmReturnExtensions);
  if(message.pixmap != (XID)0)
    XFreePixmap(dpy, message.pixmap);
  if(XpmCreatePixmapFromData(dpy, root, bytes, &message.pixmap, 
			     &message.mask, &message.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
}   


/**********************************************************************
 *
 *          Reacting to events
 *
 **********************************************************************/


/*
 * Redraw the window
 */

void RedrawWindow(void)
{
  int i=0;

  FlushExpose (theWin);
  XSetClipMask(dpy, NormalGC, wmcd.mask);
  XSetClipOrigin(dpy, NormalGC, 0, 0);
  XCopyArea(dpy, wmcd.pixmap, theWin, NormalGC,
	    0, 0, wmcd.attributes.width, wmcd.attributes.height, 0, 0);

  if(cdrom_tray_open())
    {
      trayctl = closetray;
      trayctl_p = closetray_p;
    }
  else
    {
      trayctl = opentray;
      trayctl_p = opentray_p;
    }

  XSetClipMask(dpy, NormalGC, trayctl.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[TRAYCTL], ypos[TRAYCTL]);
  XCopyArea(dpy, trayctl.pixmap, theWin, NormalGC,
	    0, 0, trayctl.attributes.width,
	    trayctl.attributes.height, xpos[TRAYCTL],
	    ypos[TRAYCTL]);

  if(!show_led || !cdrom_cd_in_drive())
    {
      XSetClipMask(dpy, NormalGC, message.mask);
      XSetClipOrigin(dpy, NormalGC, xpos[MESSAGE], ypos[MESSAGE]);
       XCopyArea(dpy, message.pixmap, theWin, NormalGC,
		0, 0, message.attributes.width,
		message.attributes.height, xpos[MESSAGE],
		ypos[MESSAGE]);
    }
  else
    DrawLED();

  XSetClipMask(dpy, NormalGC, play.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+i*CONTROLWIDTH,
                 ypos[CONTROLS]);
  XCopyArea(dpy, play.pixmap, theWin, NormalGC,
	    0, 0, play.attributes.width,
	    play.attributes.height,
	    xpos[CONTROLS]+i*CONTROLWIDTH, 
	    ypos[CONTROLS]);
  i++;

  XSetClipMask(dpy, NormalGC, stop.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+i*CONTROLWIDTH,
                 ypos[CONTROLS]); 
  XCopyArea(dpy, stop.pixmap, theWin, NormalGC,
	    0, 0, stop.attributes.width,
	    stop.attributes.height,
	    xpos[CONTROLS]+i*CONTROLWIDTH, 
	    ypos[CONTROLS]);
  i++;

  XSetClipMask(dpy, NormalGC, pause_img.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+i*CONTROLWIDTH,
                 ypos[CONTROLS]); 
  XCopyArea(dpy, pause_img.pixmap, theWin, NormalGC,
	    0, 0, pause_img.attributes.width,
	    pause_img.attributes.height,
	    xpos[CONTROLS]+i*CONTROLWIDTH,
	    ypos[CONTROLS]);
  i++;

  XSetClipMask(dpy, NormalGC, skipf.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+i*CONTROLWIDTH,
                 ypos[CONTROLS]);
  XCopyArea(dpy, skipf.pixmap, theWin, NormalGC,
	    0, 0, skipf.attributes.width,
	    skipf.attributes.height,
	    xpos[CONTROLS]+i*CONTROLWIDTH, 
	    ypos[CONTROLS]);
  i++;

  XSetClipMask(dpy, NormalGC, skipb.mask);
  XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+i*CONTROLWIDTH,
                 ypos[CONTROLS]); 
  XCopyArea(dpy, skipb.pixmap, theWin, NormalGC,
	    0, 0, skipb.attributes.width,
	    skipb.attributes.height,
	    xpos[CONTROLS]+i*CONTROLWIDTH,
	    ypos[CONTROLS]);


  if(!cdrom_tray_open() && cdrom_cd_in_drive())
    {
      if(cdrom_cd_is_data_only())
	{
          XSetClipMask(dpy, NormalGC, data.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[M_D_ICON],
                         ypos[M_D_ICON]); 
	  XCopyArea(dpy, data.pixmap, theWin, NormalGC,
		    0, 0, data.attributes.width,
		    data.attributes.height,
		    xpos[M_D_ICON],
		    ypos[M_D_ICON]);
	}
      else
	{
          XSetClipMask(dpy, NormalGC, music.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[M_D_ICON],
                         ypos[M_D_ICON]); 
	  XCopyArea(dpy, music.pixmap, theWin, NormalGC,
		    0, 0, music.attributes.width,
		    music.attributes.height,
		    xpos[M_D_ICON],
		    ypos[M_D_ICON]);
	  if(cdrom_cd_contains_data())
	    {
              XSetClipMask(dpy, NormalGC, data.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[M_D_ICON]+M_D_WIDTH+M_D_SKIP,
                             ypos[M_D_ICON]); 
	      XCopyArea(dpy, data.pixmap, theWin, NormalGC,
			0, 0, data.attributes.width,
			data.attributes.height,
			xpos[M_D_ICON]+M_D_WIDTH+M_D_SKIP,
			ypos[M_D_ICON]);
	    }	
	}
    }
  FlushExpose(theWin);
}
  

/*
 * Handle a pressed button
 */

void HandleButton(XButtonEvent *Event)
{
  int i, n;
  XEvent *e;
  int found = False;
  int grabbed = False;

  e = (XEvent *)malloc(sizeof(XEvent));

  for(i=0;i<6;i++)
    {
      if(XPointInRegion(regions[i], Event->x, Event->y))
	{
	  found=True;
	  n=i;
	  i=6;
	}
    }

  if(!found)
    {
      free(e);
      return;
    }

  switch(n)
    {
    case TRAYCTL_R:
      XSetClipMask(dpy, NormalGC, trayctl_p.mask);
      XSetClipOrigin(dpy, NormalGC, xpos[TRAYCTL], ypos[TRAYCTL]);
       XCopyArea(dpy, trayctl_p.pixmap, theWin, NormalGC,
		0, 0, trayctl_p.attributes.width,
		trayctl_p.attributes.height, xpos[TRAYCTL],
		ypos[TRAYCTL]);
      break;

    case PLAY_R:
      if(play_active)
	{
          XSetClipMask(dpy, NormalGC, play_p.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS], ypos[CONTROLS]);
	  XCopyArea(dpy, play_p.pixmap, theWin, NormalGC,
		    0, 0, play_p.attributes.width,
		    play_p.attributes.height, xpos[CONTROLS],
		    ypos[CONTROLS]);
	}
      else
	found = False;
      break;

    case STOP_R:
      if(stop_active)
	{
          XSetClipMask(dpy, NormalGC, stop_p.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+CONTROLWIDTH, ypos[CONTROLS]);
	  XCopyArea(dpy, stop_p.pixmap, theWin, NormalGC,
		    0, 0, stop_p.attributes.width,
		    stop_p.attributes.height, xpos[CONTROLS]+CONTROLWIDTH,
		    ypos[CONTROLS]);
	}
      else
	found = False;
      break;

    case PAUSE_R:
      if(pause_active)
	{
          XSetClipMask(dpy, NormalGC, pause_p.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+2*CONTROLWIDTH, ypos[CONTROLS]);
 	  XCopyArea(dpy, pause_p.pixmap, theWin, NormalGC,
		    0, 0, pause_p.attributes.width,
		    pause_p.attributes.height, xpos[CONTROLS]+2*CONTROLWIDTH,
		    ypos[CONTROLS]);
	}
      else
	found = False;
      break;

    case SKIPF_R:
      if(skipf_active)
	{
          XSetClipMask(dpy, NormalGC, skipf_p.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+3*CONTROLWIDTH, ypos[CONTROLS]);
 	  XCopyArea(dpy, skipf_p.pixmap, theWin, NormalGC,
		    0, 0, skipf_p.attributes.width,
		    skipf_p.attributes.height, xpos[CONTROLS]+3*CONTROLWIDTH,
		    ypos[CONTROLS]);

	  if(Event->button == Button3) 
	    {
	      int skipping = True;
	      int i, max, skipped;
	      cdrom_position_t absolute, pos, track;
	      

 	      XGrabPointer(dpy, theWin, True, ButtonReleaseMask,
 			   GrabModeAsync, GrabModeAsync, None, None,
 			   CurrentTime); 
 	      while(skipping)
 		{
 		  Ffwd();
 		  if(XCheckMaskEvent(dpy, ButtonReleaseMask, e))
 		    {
 		      if(((XButtonEvent *)e)->button == Button3)
 			skipping = False;
 		    }
 		  else usleep(100000);
 		}
              XSetClipMask(dpy, NormalGC, skipf.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+3*CONTROLWIDTH,
                             ypos[CONTROLS]); 
	      XCopyArea(dpy, skipf.pixmap, theWin, NormalGC,
			0, 0, skipf.attributes.width,
			skipf.attributes.height, xpos[CONTROLS]+3*CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      free(e);
	      return;
	    }
		  
	}
      else
	found = False;
      break;

    case SKIPB_R:
      if(pause_active)
	{
          XSetClipMask(dpy, NormalGC, skipb_p.mask);
          XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+4*CONTROLWIDTH, ypos[CONTROLS]);
 	  XCopyArea(dpy, skipb_p.pixmap, theWin, NormalGC,
		    0, 0, skipb_p.attributes.width,
		    skipb_p.attributes.height, xpos[CONTROLS]+4*CONTROLWIDTH,
		    ypos[CONTROLS]);

	  /* Reverse currently disabled 'cos it don't work. See
	     BUGS. */
	  /*	  if(Event->button == Button3) 
	    {
	      int skipping = True;
	      
	      XGrabPointer(dpy, theWin, True, ButtonReleaseMask,
			   GrabModeAsync, GrabModeAsync, None, None,
			   CurrentTime); 
	      while(skipping)
		{
		  Rev();
		  if(XCheckMaskEvent(dpy, ButtonReleaseMask, e))
		    {
		      if(((XButtonEvent *)e)->button == Button3)
			skipping = False;
		    }
		  else usleep(100000);
		}*/
              XSetClipMask(dpy, NormalGC, skipf.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+3*CONTROLWIDTH,
                             ypos[CONTROLS]); 
	      XCopyArea(dpy, skipf.pixmap, theWin, NormalGC,
			0, 0, skipf.attributes.width,
			skipf.attributes.height, xpos[CONTROLS]+3*CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      free(e);
	      return;
	    }
      else
	found = False;
      break;
      
    default:
      printf("Pressed button somewhere\n");
      found = False;
      break;
    }

  if(!found)
    {
      free(e);
      return;
    }

  XGrabPointer(dpy, theWin, True, ButtonReleaseMask, GrabModeAsync,
	       GrabModeAsync, None, None, CurrentTime);
  grabbed = True;

  while(grabbed)
    {
      XNextEvent(dpy, e);
      
      if(e->type == ButtonRelease)
	{
	  switch(n)
	    {
	    case TRAYCTL_R:
              XSetClipMask(dpy, NormalGC, trayctl.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[TRAYCTL],
                             ypos[TRAYCTL]); 
	      XCopyArea(dpy, trayctl.pixmap, theWin, NormalGC,
			0, 0, trayctl.attributes.width,
			trayctl.attributes.height, xpos[TRAYCTL],
			ypos[TRAYCTL]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[TRAYCTL_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		if((cdrom_cd_contains_data() || cdrom_tray_open()) && ((XButtonEvent *)e)->button == Button3)
		  TrayMount();
		else
		  TrayCtl();
	      
	      break;
	    case PLAY_R:
              XSetClipMask(dpy, NormalGC, play.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS],
                             ypos[CONTROLS]); 
	      XCopyArea(dpy, play.pixmap, theWin, NormalGC,
			0, 0, play.attributes.width,
			play.attributes.height, xpos[CONTROLS],
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[PLAY_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		  Play();
	      break;
	    case STOP_R:
              XSetClipMask(dpy, NormalGC, stop.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+CONTROLWIDTH,
                             ypos[CONTROLS]); 
 	      XCopyArea(dpy, stop.pixmap, theWin, NormalGC,
			0, 0, stop.attributes.width,
			stop.attributes.height, xpos[CONTROLS]+CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[STOP_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		Stop();
	      break;
	    case PAUSE_R:
              XSetClipMask(dpy, NormalGC, pause_img.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+2*CONTROLWIDTH,
                             ypos[CONTROLS]); 
 	      XCopyArea(dpy, pause_img.pixmap, theWin, NormalGC,
			0, 0, pause_img.attributes.width,
			pause_img.attributes.height, xpos[CONTROLS]+2*CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[PAUSE_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		Pause();
	      break;
	    case SKIPF_R:
              XSetClipMask(dpy, NormalGC, skipf.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+3*CONTROLWIDTH,
                             ypos[CONTROLS]); 
	      XCopyArea(dpy, skipf.pixmap, theWin, NormalGC,
			0, 0, skipf.attributes.width,
			skipf.attributes.height, xpos[CONTROLS]+3*CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy);
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[SKIPF_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		SkipF();
	      break;
	    case SKIPB_R:
              XSetClipMask(dpy, NormalGC, skipb.mask);
              XSetClipOrigin(dpy, NormalGC, xpos[CONTROLS]+4*CONTROLWIDTH,
                             ypos[CONTROLS]); 
 	      XCopyArea(dpy, skipb.pixmap, theWin, NormalGC,
			0, 0, skipb.attributes.width,
			skipb.attributes.height, xpos[CONTROLS]+4*CONTROLWIDTH,
			ypos[CONTROLS]);
	      XUngrabPointer(dpy, CurrentTime);
	      XFlush(dpy); 
	      grabbed = False;

	      if(((XButtonEvent*)e)->window == theWin &&
		 XPointInRegion(regions[SKIPB_R],
				((XButtonEvent*)e)->x,
				((XButtonEvent*)e)->y)) 
		SkipB();
	      break;
	    default:
	    }
	}
    }
  free(e);
}


/*
 * The main run loop
 */

void RunLoop(void)
{
  XEvent Event;
  int ctr;
  
  ctr = 0;
  
  while(1)
    {
      XEventsQueued(dpy, QueuedAfterReading);
      if(XCheckMaskEvent(dpy, ButtonPressMask|ExposureMask, &Event))
	switch(Event.type)
	  {
	  case ButtonPress:
	    HandleButton((XButtonEvent *)&Event);
	  case Expose:
	    RedrawWindow();
	  default:
	  }

      /* redraw the window to update the timer display */
      if(!cdrom_mounted())
	{
	  if(cdrom_cd_is_data_only() || !cdrom_cd_in_drive())
	    {
	      if(l_mounted)
		{
		  CreateMessage(startup_xpm);
		  RedrawWindow();
		}
	    }
	  else
	    {
	      show_led = True;
	      if(l_mounted)
		{
		  UpdateShape();
		  RedrawWindow();
		}
	      if(ReadPosition())
		DrawLED();
	    }
	  l_mounted = False;
	}
      else
	{
	  if(!l_mounted)
	    {
	      EvalState();
	      CreateMessage(mounted_xpm);
	      show_led = False;
	      UpdateShape();
	      RedrawWindow();
	    }
	  l_mounted = True;
	}
  
      /* sleep 1/10 sec to keep the load down */
      usleep(100000);
      ctr++;
#ifndef CHECKINTERVAL
      if(ctr==10)
#else
      if(ctr==CHECKINTERVAL)
#endif
	{
	  ctr=0;
#ifndef NOUPDATE
	  cdrom_update_state();
#endif
	}
    }
}


/**********************************************************************
 *
 *          Setup
 *
 **********************************************************************/


/*
 * Setup pixmaps
 */

void GetXPM(void)
{
  wmcd.attributes.valuemask |=(XpmReturnPixels | XpmReturnExtensions);
  if(shaped)
    {
      if(XpmCreatePixmapFromData(dpy, root, wmcd_shaped_xpm, &wmcd.pixmap,
				 &wmcd.mask, &wmcd.attributes) 
	 != XpmSuccess)
	{
	  fprintf(stderr, ERR_colorcells);
	  exit(1);
	}
    }
  else
    {
      wmcd.attributes.valuemask |=(XpmReturnPixels | XpmReturnExtensions);
      if(XpmCreatePixmapFromData(dpy, root, wmcd_xpm, &wmcd.pixmap,
				 &wmcd.mask, &wmcd.attributes) 
	 != XpmSuccess)
	{
	  fprintf(stderr, ERR_colorcells);
	  exit(1);
	}
    }
  led.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, led_xpm, &led.pixmap, 
			     &led.mask, &led.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  /* The Actuators: The active... */

  closetray.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, closetray_xpm, &closetray.pixmap, 
			     &closetray.mask, &closetray.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
  
  opentray.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, opentray_xpm, &opentray.pixmap, 
			     &opentray.mask, &opentray.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }

  play_a.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, play_xpm, &play_a.pixmap, 
			     &play_a.mask, &play_a.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
  
  stop_a.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, stop_xpm, &stop_a.pixmap, 
			     &stop_a.mask, &stop_a.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  pause_a.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, pause_xpm, &pause_a.pixmap, 
			     &pause_a.mask, &pause_a.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  skipf_a.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipf_xpm, &skipf_a.pixmap, 
			     &skipf_a.mask, &skipf_a.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
  
  skipb_a.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipb_xpm, &skipb_a.pixmap, 
			     &skipb_a.mask, &skipb_a.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  /* ..., the inactive... */

  play_i.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, play_i_xpm, &play_i.pixmap, 
			     &play_i.mask, &play_i.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  stop_i.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, stop_i_xpm, &stop_i.pixmap, 
			     &stop_i.mask, &stop_i.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  pause_i.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, pause_i_xpm, &pause_i.pixmap, 
			     &pause_i.mask, &pause_i.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  skipf_i.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipf_i_xpm, &skipf_i.pixmap, 
			     &skipf_i.mask, &skipf_i.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  skipb_i.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipb_i_xpm, &skipb_i.pixmap, 
			     &skipb_i.mask, &skipb_i.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  /* ... and the pushed variety */

  opentray_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, opentray_p_xpm, &opentray_p.pixmap, 
			     &opentray_p.mask, &opentray_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }

  closetray_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, closetray_p_xpm, &closetray_p.pixmap, 
			     &closetray_p.mask, &closetray_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  play_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, play_p_xpm, &play_p.pixmap, 
			     &play_p.mask, &play_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  stop_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, stop_p_xpm, &stop_p.pixmap, 
			     &stop_p.mask, &stop_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  pause_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, pause_p_xpm, &pause_p.pixmap, 
			     &pause_p.mask, &pause_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  skipf_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipf_p_xpm, &skipf_p.pixmap, 
			     &skipf_p.mask, &skipf_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
    
  skipb_p.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, skipb_p_xpm, &skipb_p.pixmap, 
			     &skipb_p.mask, &skipb_p.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }

  /* The music/data icons */

  music.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, music_xpm, &music.pixmap, 
			     &music.mask, &music.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }
  data.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  if(XpmCreatePixmapFromData(dpy, root, data_xpm, &data.pixmap, 
			     &data.mask, &data.attributes)
     != XpmSuccess)
    {
      fprintf(stderr, ERR_colorcells);
      exit(1);
    }

  message.pixmap = (XID)0;
}


/*
 * Setup basic shape.
 */

void SetShape(void)
{
  if(!shaped) return;
  XShapeCombineMask(dpy, theWin, ShapeBounding, 0, 0, wmcd.mask,
		    ShapeSet);
}

/*
 * Update shape
 */

void UpdateShape(void)
{
  int i=0;

  if(!shaped) return;
  
  if(play_active && !l_play)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			play_a.mask, ShapeUnion);
      l_play = True;
    }
  else if(!play_active && l_play)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			play_a.mask, ShapeSubtract);
      l_play = False;
    }

  i++;

  if(stop_active && !l_stop)
    {
      
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			stop_a.mask, ShapeUnion);
      l_stop = True;
    }
  else if(!stop_active && l_stop)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			stop_a.mask, ShapeSubtract);
      l_stop = False;
    }

  i++;

  if(pause_active && !l_pause)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			pause_a.mask, ShapeUnion);
      l_pause = True;
    }
  else if(!pause_active && l_pause)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			pause_a.mask, ShapeSubtract);
      l_pause = False;
    }
  
  i++;
  
  if(skipf_active && !l_skipf)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			skipf_a.mask, ShapeUnion);
      l_skipf = True;
    }
  else if(!skipf_active && l_skipf)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			skipf_a.mask, ShapeSubtract);
      l_skipf = False;
    }
  
  i++;
  
  if(skipb_active && !l_skipb)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			skipb_a.mask, ShapeUnion);
      l_skipb = True;
    }
  else if(!skipb_active && l_skipb)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[CONTROLS]+i*CONTROLWIDTH, ypos[CONTROLS],
			skipb_a.mask, ShapeSubtract);
      l_skipb = False;
    }
      
  if(cdrom_tray_open() && !l_tray)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[TRAYCTL], ypos[TRAYCTL],
			opentray.mask, ShapeSubtract);
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[TRAYCTL], ypos[TRAYCTL],
			closetray.mask, ShapeUnion);
      l_tray = True;
    }
  else if(!cdrom_tray_open() && l_tray)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[TRAYCTL], ypos[TRAYCTL],
			closetray.mask, ShapeSubtract);
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[TRAYCTL], ypos[TRAYCTL],
			opentray.mask, ShapeUnion);
      l_tray = False;
    }

  if(l_music)
    {
      XShapeCombineMask(dpy, theWin, ShapeBounding,
			xpos[M_D_ICON], ypos[M_D_ICON], music.mask,
			ShapeSubtract);
      if(l_data)
	XShapeCombineMask(dpy, theWin, ShapeBounding,
			  xpos[M_D_ICON]+M_D_WIDTH+M_D_SKIP,
			  ypos[M_D_ICON],
			  data.mask, ShapeSubtract);
    }
  else if(l_data)
    XShapeCombineMask(dpy, theWin, ShapeBounding,
		      xpos[M_D_ICON], ypos[M_D_ICON], data.mask,
		      ShapeSubtract);
			  
  
  if(!cdrom_tray_open() && cdrom_cd_in_drive())
    {
      if(cdrom_cd_is_data_only())
	{
	  XShapeCombineMask(dpy, theWin, ShapeBounding,
			    xpos[M_D_ICON], ypos[M_D_ICON], data.mask,
			    ShapeUnion);
	  l_data = True;
	}
      else
	{
	  XShapeCombineMask(dpy, theWin, ShapeBounding,
			    xpos[M_D_ICON], ypos[M_D_ICON],
			    music.mask, ShapeUnion);
	  l_music = True;
	  if(cdrom_cd_contains_data())
	    {
	      XShapeCombineMask(dpy, theWin, ShapeBounding,
				xpos[M_D_ICON]+M_D_WIDTH+M_D_SKIP,
				ypos[M_D_ICON],
				data.mask, ShapeUnion);
	      l_data = True;
	    }
	}
    }
  else
    {
      l_music = False; l_data = False;
    }
  
}


/*
 * Setup regions. These are important when determining whether a button
 * has been pressed.
 */

void CreateRegions(void)
{
  int i;
  XPoint points[5];

  if(cdrom_tray_open())
    trayctl = closetray;
  else
    trayctl = opentray;

  points[0].x = xpos[TRAYCTL];
  points[0].y = ypos[TRAYCTL];
  points[1].x = xpos[TRAYCTL] + trayctl.attributes.width;
  points[1].y = ypos[TRAYCTL];
  points[2].x = xpos[TRAYCTL] + trayctl.attributes.width;
  points[2].y = xpos[TRAYCTL] + trayctl.attributes.height;
  points[3].x = xpos[TRAYCTL];
  points[3].y = xpos[TRAYCTL] + trayctl.attributes.height;
  points[4].x = xpos[TRAYCTL];
  points[4].y = ypos[TRAYCTL];
  regions[TRAYCTL_R] = XPolygonRegion(points, 5, WindingRule);

  for(i=0;i<5;i++)
    {
      points[0].x = xpos[CONTROLS]+i*CONTROLWIDTH;
      points[0].y = ypos[CONTROLS];
      points[1].x = xpos[CONTROLS]+(i+1)*CONTROLWIDTH;
      points[1].y = ypos[CONTROLS];
      points[2].x = xpos[CONTROLS]+(i+1)*CONTROLWIDTH;
      points[2].y = ypos[CONTROLS]+CONTROLHEIGHT;
      points[3].x = xpos[CONTROLS]+i*CONTROLWIDTH;
      points[3].y = ypos[CONTROLS]+CONTROLHEIGHT;
      points[4].x = xpos[CONTROLS]+i*CONTROLWIDTH;
      points[4].y = ypos[CONTROLS];
      regions[i+1] = XPolygonRegion(points, 5, WindingRule);
    }
}


/*
 * Setup the user interface
 */

void InitUI(int argc, char **argv)
{
  XGCValues gcv;
  unsigned long gcm;
  char *display_name = NULL;
  XSizeHints sizehints;
  XWMHints wmhints;
  XClassHint classhint;
  char *wname = "wmcd";
  char *client_machine;
  XTextProperty name;
  XTextProperty machine;
  GNUstepWMAttributes gsattrs;
  Atom _XA_GNUSTEP_WM_ATTR;
  int i;

  gsattrs.flags = GSWindowStyleAttr | GSWindowLevelAttr;
  gsattrs.window_level = NSNormalWindowLevel;
  gsattrs.window_style = NSTitledWindowMask | NSClosableWindowMask |
    NSMiniaturizableWindowMask; 

  for(i=1; i<argc; i++)
    {
      if(!strcmp(argv[i], "-noshape"))
	shaped = False;
      else if(!strcmp(argv[i], "-noticonic"))
	iconic = False;
      else if(!strcmp(argv[i], "-notitle"))
	gsattrs.window_style = NSBorderlessWindowMask;
      else if(!strcmp(argv[i], "-ontop"))
	gsattrs.window_level = NSFloatingWindowLevel;
      else if(!strcmp(argv[i], "-device"))
	i++;
      else if(!strcmp(argv[i], "-nomount")); /* do nothing */
      else
	{
	  fprintf(stderr, "wmcd: Don't know option %s\n", argv[i]);
	  fprintf(stderr, "Usage: wmcd [-device DEVICE] [-nomount]\n"
		          "            [-noshape] [-noticonic [-ontop] [-notitle]]\n");
	  exit(1);
	}
    }

  if(!(dpy = XOpenDisplay(display_name)))
    {
      fprintf(stderr, "wmcd: Can't open Display %s\n", XDisplayName(display_name));
      exit(1);
    }

  screen = DefaultScreen(dpy);
  root = RootWindow(dpy, screen);
  d_depth = DefaultDepth(dpy, screen);
  stopwatch = XCreateFontCursor(dpy, XC_watch);

  GetXPM();
  CreateRegions();

  CreateMessage(startup_xpm);
  show_led = False;
  
  back_pix = GetColor("white");
  fore_pix = GetColor("#282828");

  sizehints.flags = USSize | PMinSize | PMaxSize;
  sizehints.width = wmcd.attributes.width;
  sizehints.height = wmcd.attributes.height;
  sizehints.min_width = wmcd.attributes.width;
  sizehints.min_height = wmcd.attributes.height;
  sizehints.max_width = wmcd.attributes.width;
  sizehints.max_height = wmcd.attributes.height;
  
  win = XCreateSimpleWindow(dpy, root, 0, 0,
			    sizehints.width, sizehints.height,
			    0, fore_pix, back_pix);
  if(iconic)
    {
      iconwin = XCreateSimpleWindow(dpy, win, 0, 0,
				    sizehints.width, sizehints.height,
				    0, fore_pix, back_pix);
      theWin = iconwin;
    }
  else
    theWin = win;

  XSetWMNormalHints(dpy, win, &sizehints);

  classhint.res_name =  "wmcd";
  classhint.res_class = "WMcd";
  XSetClassHint(dpy, win, &classhint);
  if(iconic) XSetClassHint(dpy, iconwin, &classhint);

  if (XStringListToTextProperty(&wname, 1, &name) ==0)
    {
      fprintf(stderr, "wmcd: can't allocate window name\n");
      exit(-1);
    }

  XSetWMName(dpy, win, &name);

  client_machine = (char *)malloc(256);
  if(gethostname(client_machine, 255)>=0)
    {
      if(XStringListToTextProperty(&client_machine, 1, &machine) != 0)
	XSetWMClientMachine(dpy, win, &machine);
      else
	fprintf(stderr, "wmcd: can't allocate client machine name\n");
    }
  else
    {
      fprintf(stderr, "wmcd: can't get hostname\n");
      perror("wmcd: gethostname()");
    }
  
  if(iconic) XSetWMName(dpy, iconwin, &name);

  XSelectInput(dpy, win, MW_EVENTS);
  if(iconic) XSelectInput(dpy, iconwin, MW_EVENTS);
  
  gcm = GCForeground|GCBackground|GCGraphicsExposures;
  gcv.foreground = fore_pix;
  gcv.background = back_pix;
  gcv.graphics_exposures = False;
  NormalGC = XCreateGC(dpy, root, gcm, &gcv);  

  if(iconic)
    {
      wmhints.initial_state = (WithdrawnState);
      wmhints.icon_window = iconwin;
    }
  else
    wmhints.initial_state = NormalState;
  
  wmhints.icon_x = sizehints.x;
  wmhints.icon_y = sizehints.y;
  wmhints.window_group = win;
  wmhints.flags = StateHint |
    (iconic ? IconWindowHint : 0x0) | IconPositionHint |
    WindowGroupHint;
  XSetWMHints(dpy, win, &wmhints); 
  
  XSetCommand(dpy, win, argv, argc);
  
  _XA_GNUSTEP_WM_ATTR = XInternAtom(dpy, "_GNUSTEP_WM_ATTR", False);
  XChangeProperty(dpy, win, _XA_GNUSTEP_WM_ATTR, _XA_GNUSTEP_WM_ATTR,
		  32, PropModeReplace, (unsigned char *)&gsattrs,
		  sizeof(GNUstepWMAttributes)/sizeof(CARD32));
  
  EvalState();
  
  l_tray = !cdrom_tray_open();
  
  SetShape();
  UpdateShape();
  
  XMapWindow(dpy,win);
  
}


