
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <linux/fs.h>
#include <sys/mount.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "cdrom.h"

#include "config.h"

#include "typedefs.h"
#include "window.h"


/* Globale Variablen */

char *ERR_colorcells = "Not enough free color cells\n";

/* State Variables */

int show_led;

extern XpmIcon play, play_a, play_i, play_p;
extern XpmIcon stop, stop_a, stop_i, stop_p;
extern XpmIcon pause_img, pause_a, pause_i, pause_p;
extern XpmIcon skipf, skipf_a, skipf_i, skipf_p;
extern XpmIcon skipb, skipb_a, skipb_i, skipb_p;

int play_active = False;
int stop_active = False;
int pause_active = False;
int skipf_active = False;
int skipb_active = False;


int track_playing = -1;
cdrom_position_t position;

void EvalState(void)
{
  if(cdrom_tray_open())
    {
      play = play_a;
      play_active = True;
      stop = stop_i;
      stop_active = False;
      pause_img = pause_i;
      pause_active = False;
      skipf = skipf_i;
      skipf_active = False;
      skipb = skipb_i;
      skipb_active = False;
    }
  else
    {
      if(cdrom_cd_in_drive())
	{
	  if(cdrom_mounted() || cdrom_cd_is_data_only())
	    {
	      if(cdrom_mounted() && !cdrom_cd_is_data_only() && cdrom_cd_contains_data())
		{
		  play = play_a;
		  play_active = True;
		}
	      else
		{
		  play = play_i;
		  play_active = False;
		}
	      stop = stop_i;
	      stop_active = False;
	      pause_img = pause_i;
	      pause_active = False;
	      skipf = skipf_i;
	      skipf_active = False;
	      skipb = skipb_i;
	      skipf_active = False;
	    }
	  else /* if(!(cdrom_mounted() || cdrom_cd_is_data_only)) */
	    {
	      play = play_a;
	      play_active = True;
	      stop = stop_a;
	      stop_active = True;
	      pause_img = pause_a;
	      pause_active = True;
	      skipf = skipf_a;
	      skipf_active = True;
	      skipb = skipb_a;
	      skipb_active = True;

	      show_led = True;
	    }
	} 
      else /* if(!cdrom_cd_in_drive()) */
	{
	  play = play_i;
	  play_active = False;
	  stop = stop_i;
	  stop_active = False;
	  pause_img = pause_i;
	  pause_active = False;
	  skipf = skipf_i;
	  skipf_active = False;
	  skipb = skipb_i;
	  skipb_active = False;
	}
    }
}

  
int ReadPosition(void)
{
  int ret = 0;
  cdrom_position_t new_pos;
  if (cdrom_get_position_relative(&new_pos) == -1 ) {
    fprintf(stderr, "ioctl cdromsubchnl\n");
    exit(1);
  }
  if((new_pos.minute != position.minute ||
      new_pos.second != position.second) ||
     (cdrom_playing() && cdrom_current_track() != track_playing))
    {
      position.minute = new_pos.minute;
      position.second = new_pos.second;
      ret = 1;
    }
  
  if(cdrom_playing()) track_playing = cdrom_current_track();
  
  return ret;
}


int main(int argc, char *argv[])
{
  char *dev;
  BOOL mount;
  int i;

  mount = True;
  dev = DEVICE;

  for(i=1; i<argc; i++)
    if(!strcmp(argv[i], "-device"))
      dev = argv[++i];
    else if(!strcmp(argv[i], "-nomount"))
      mount = False;
  
  if(init_cdrom(dev, mount) == -1)
    {
      fprintf(stderr, "Couldn't initialize CDROM:\n");
      exit(1);
    }
  
  InitUI(argc, argv);

  if(cdrom_update_state() == -1)
    fprintf(stderr, "Error while updating CDROM state\n");

  RunLoop();

  exit(0);
}


